#include <netinet/in.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <linux/if.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

#include "monarch.h"
#include "proxies.h"
#include "probes.h"
#include "traces.h"

#define dprintf(a...) do { } while (0)

PlanetLabProxy::PlanetLabProxy() : Proxy()
{
  /* Initialize variables. Do not open any files or sockets! */
  this->networkDeviceName[0] = 0;
  this->initialized = false;
  this->tapHandle = -1;
  this->packetTrace = NULL;
  this->captureSocket = -1;
  this->sendSocket = -1;
  memset(&addresses, 0, sizeof(addresses));
}

PlanetLabProxy::~PlanetLabProxy()
{
  /* Close any open sockets, release any allocated resources 
     Note: May be called at any time (e.g. from the SIGINT handler),
     so resources may be partially allocated */

  if (captureSocket >= 0) {
    close(captureSocket);
    captureSocket = -1;
  }
  
  if (sendSocket >= 0) {
    close(sendSocket);
    sendSocket = -1;
  }
  
  if (tapHandle >= 0) {
    close(tapHandle);
    tapHandle = -1;
  }

}


void PlanetLabProxy::setupTAP(struct addressBlock *addressArg)
{
  /* All the serious initialization stuff (files, sockets, ...) */

  char buf1[256], buf2[256];
  assert(tapHandle < 0);

  struct ifreq ifr;
  struct sockaddr_in *saddr;
  int sd; // socket for configuring the tap device

  /* Regarding /dev/net/tun: Should be present on every PlanetLab node. But what we get back is in fact a tap device handle and no tun device (this means, we get ethernet frames instead of IP packets) */
  if ((tapHandle = open("/dev/net/tun", O_RDWR)) < 0) 
    panic(ERR_NOTFOUND, "PlanetLabProxy: Cannot open /dev/net/tun (code %d). Is this really an up-to-date PlanetLab host?", tapHandle);

  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = IFF_TAP | IFF_NO_PI; /* TAP device; do not provide packet information */

  if (ioctl(tapHandle, TUNSETIFF, (void*) &ifr) < 0)
    panic(ERR_INTERNAL, "PlanetLabProxy: Cannot set TAP device interface (%s)", strerror(errno));

  strcpy(tapDeviceName, ifr.ifr_name);
  dprintf("Allocated tap device %s\n", tapDeviceName);
  int tap_num = atoi(&tapDeviceName[3]); // get number of the tap device to use it for IP address choice

  if ((sd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    panic(ERR_INTERNAL, "PlanetLabProxy: Cannot allocate socket (%s)", strerror(errno));

  /* Get the current flags */
  if (ioctl(sd, SIOCGIFADDR, &ifr) == -1) 
    panic(ERR_INTERNAL, "PlanetLabProxy: Failed to get IP address of device %s (%s)", tapDeviceName, strerror(errno));

  struct sockaddr_in tapLocalAddress;
  memcpy((char *) &tapLocalAddress, (char *) &ifr.ifr_addr, sizeof(tapLocalAddress));


  /* Get hardware address of the TAP device */
  if (ioctl(sd, SIOCGIFHWADDR, &ifr) < 0)
    panic(ERR_INTERNAL, "PlanetLabProxy: Cannot get MAC address for '%s' (%s)", tapDeviceName, strerror(errno));

  memcpy(&tapHwAddress, &ifr.ifr_hwaddr.sa_data, 6);

    
  /* Get the netmask */
  if (ioctl(sd, SIOCGIFNETMASK, &ifr) == -1) 
    panic(ERR_INTERNAL, "PlanetLabProxy: Failed to get netmask of device %s (%s)", tapDeviceName, strerror(errno));

  struct sockaddr_in tapNetmask;
  memcpy((char *) &tapNetmask, (char *) &ifr.ifr_addr, sizeof(tapNetmask));


  /* Peers */
  
  addresses = *addressArg;
  uint16 ports[2];
  findUnusedPorts(tapLocalAddress.sin_addr.s_addr, ports, 2);
  

  if (!addresses.senderAddress.sin_family) {
    addresses.senderAddress = tapLocalAddress;
    addresses.senderAddress.sin_port = ports[0];
  }
  else if(addresses.senderAddress.sin_addr.s_addr != tapLocalAddress.sin_addr.s_addr) {
    panic(ERR_PARAM, "PlanetLabProxy: On Planetlab, we cannot change the address of the tap device. This site uses %s\n                as the tap device's address while you try to bind to %s. This will not work.\n                Use the fixed address of the tap device for --sender and --receiver instead.", renderIP(ntohl(tapLocalAddress.sin_addr.s_addr), buf1), renderIP(ntohl(addresses.senderAddress.sin_addr.s_addr), buf2));
  }
  
  if (!addresses.receiverAddress.sin_family) {
    addresses.receiverAddress = tapLocalAddress;
    addresses.receiverAddress.sin_port = ports[1];
  }
  else if(addresses.receiverAddress.sin_addr.s_addr != tapLocalAddress.sin_addr.s_addr) {
    panic(ERR_PARAM, "PlanetLabProxy: On Planetlab, we cannot change the address of the tap device. This site uses %s\n                as the tap device's address while you try to bind to %s. This will not work.\n                Use the fixed address of the tap device for --sender and --receiver instead.", renderIP(ntohl(tapLocalAddress.sin_addr.s_addr), buf1), renderIP(ntohl(addresses.receiverAddress.sin_addr.s_addr), buf2));
  }
  
  if (!addresses.senderPeerAddress.sin_family) {
    addresses.senderPeerAddress = tapLocalAddress;
    addresses.senderPeerAddress.sin_addr.s_addr = htonl(ntohl(addresses.senderPeerAddress.sin_addr.s_addr) + 1);
    addresses.senderPeerAddress.sin_port = htons(10001);
  }
  else if((addresses.senderPeerAddress.sin_addr.s_addr & tapNetmask.sin_addr.s_addr) != (tapLocalAddress.sin_addr.s_addr & tapNetmask.sin_addr.s_addr)){
    char buf3[256];
    panic(ERR_PARAM, "PlanetLabProxy: On Planetlab, we cannot add any entries to the routing table.Assigning %s as the\n                address of the virtual-sender or virtual-receiver will not work. You have to choose\n                the addresses such that they are in the same subnet as the tap device (IP %s\n                and netmask %s)", renderIP(ntohl(addresses.senderPeerAddress.sin_addr.s_addr), buf1), renderIP(ntohl(tapLocalAddress.sin_addr.s_addr), buf2), renderIP(ntohl(tapNetmask.sin_addr.s_addr), buf3));
  }
  
  if (!addresses.receiverPeerAddress.sin_family) {
    addresses.receiverPeerAddress = tapLocalAddress;
    addresses.receiverPeerAddress.sin_addr.s_addr = htonl(ntohl(addresses.receiverPeerAddress.sin_addr.s_addr) + 1);
    addresses.receiverPeerAddress.sin_port = htons(10002);
  }
  else if((addresses.senderPeerAddress.sin_addr.s_addr & tapNetmask.sin_addr.s_addr) != (tapLocalAddress.sin_addr.s_addr & tapNetmask.sin_addr.s_addr)){
    char buf3[256];
    panic(ERR_PARAM, "PlanetLabProxy: On Planetlab, we cannot add any entries to the routing table.Assigning %s as the\n                address of the virtual-sender or virtual-receiver will not work. You have to choose\n                the addresses such that they are in the same subnet as the tap device (IP %s\n                and netmask %s)", renderIP(ntohl(addresses.receiverPeerAddress.sin_addr.s_addr), buf1), renderIP(ntohl(tapLocalAddress.sin_addr.s_addr), buf2), renderIP(ntohl(tapNetmask.sin_addr.s_addr), buf3));
  }
  
}

void PlanetLabProxy::init(struct addressBlock *addressArg, const char *deviceName)
{
  char buf1[256];
  struct sockaddr_in sin;
  assert(!initialized && policy && addressArg->localAddress.sin_family && addressArg->remoteAddress.sin_family);
  
  if (geteuid() != 0)
    panic(ERR_PRIVILEGE, "PlanetLabProxy: Must be run as user 'root'");
  
  strncpy(networkDeviceName, deviceName, sizeof(networkDeviceName));
  policy->setProxy(this);
  setupTAP(addressArg);
  
  /* Open raw/packet sockets */

  if ((captureSocket = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) < 0)
    panic(ERR_NOTFOUND, "PlanetLabProxy: Cannot open capture socket (%s)", strerror(errno));

  /* We need to tell Planetlab what we intend to send over a raw socket */
  if(strcmp(policy->getName(), "tcp-ack") == 0) {
    if ((sendSocket = socket(AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
      panic(ERR_NOTFOUND, "PlanetLabProxy: Cannot open send socket (%s)", strerror(errno));
  }
  else if(strcmp(policy->getName(), "udp-data") == 0) {
    if ((sendSocket = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0)
      panic(ERR_NOTFOUND, "PlanetLabProxy: Cannot open send socket (%s)", strerror(errno));
  }
  else if((strcmp(policy->getName(), "icmp-echo") == 0) || (strcmp(policy->getName(), "icmp-timestamp") == 0)) {
    if ((sendSocket = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0)
      panic(ERR_NOTFOUND, "PlanetLabProxy: Cannot open send socket (%s)", strerror(errno));
  }

  /* For PlanetLab, we have to bind the raw socket */
  memset(&sin, 0, sizeof(sin));
  memcpy((char *) &sin.sin_addr.s_addr, (char *) &addresses.localAddress.sin_addr, 4);
  sin.sin_port = addresses.localAddress.sin_port;
  sin.sin_family = AF_INET;
  if (bind(sendSocket, (struct sockaddr *) &sin, sizeof(sin)) < 0){
    panic(ERR_CANTBIND, "Cannot bind the send socket (%s)", strerror(errno));
  }

  /* Get interface index for DEV */
  struct ifreq ifr;
  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, networkDeviceName, sizeof(ifr.ifr_name));
  if (ioctl(captureSocket, SIOCGIFINDEX, &ifr) < 0)
    panic(ERR_INTERNAL, "TunProxy: Cannot get interface index for '%s' (%s)", networkDeviceName, strerror(errno));

  /* Bind capture socket to DEV */
  struct sockaddr_ll ll;
  memset(&ll, 0, sizeof(ll));
  ll.sll_family = AF_PACKET;
  ll.sll_ifindex = ifr.ifr_ifindex;
  ll.sll_protocol = htons(ETH_P_IP);
  if (bind(captureSocket, (struct sockaddr*)&ll, sizeof(ll)) < 0)
    panic(ERR_CANTBIND, "Cannot bind capture socket (%s)", strerror(errno));

  /* Set HDRINCL option on send socket */
  int flag = 1;
  if (setsockopt(sendSocket, IPPROTO_IP, IP_HDRINCL, (char*)&flag, sizeof(flag)) < 0)
    panic(ERR_INTERNAL, "Cannot set IP_HDRINCL option on send socket (%s)", strerror(errno));

  /* For PlanetLab we have to vary the destination port, so remember the last one we used */
  lastport = ntohs(addresses.remoteAddress.sin_port);

  /* Print some statistics to the trace file */
  if (packetTrace) {
    packetTrace->addComment("Local IP: %s", renderIP(ntohl(addresses.localAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("Remote IP: %s", renderIP(ntohl(addresses.remoteAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("TUN IP: %s", renderIP(ntohl(addresses.senderAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("FAKE IP: %s", renderIP(ntohl(addresses.senderPeerAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("Sender Port %d Receiver Port %d Proxy Port %d Remote Port %d", ntohs(addresses.senderAddress.sin_port), ntohs(addresses.receiverAddress.sin_port), ntohs(addresses.localAddress.sin_port), ntohs(addresses.remoteAddress.sin_port));
  }

  initialized = true;
}

void PlanetLabProxy::getAddresses(struct sockaddr_in *senderLocal, struct sockaddr_in *senderRemote, struct sockaddr_in *receiverLocal, struct sockaddr_in *receiverRemote)
{
  /* Tell the sender and the receiver which addresses to use */
  if (senderLocal)
    *senderLocal = addresses.senderAddress;
  if (senderRemote)
    *senderRemote = addresses.senderPeerAddress;
  if (receiverLocal)
    *receiverLocal = addresses.receiverAddress;
  if (receiverRemote)
    *receiverRemote = addresses.receiverPeerAddress;
}

void PlanetLabProxy::sendRaw(Packet *p) 
{
  struct sockaddr_in sockAddr;
  int nbytes = 0;
  sockAddr.sin_family = AF_INET;
  sockAddr.sin_addr.s_addr = htonl(p->getDestination());
  sockAddr.sin_port = htons(p->getDestinationPort());
  struct timeval tv;
  gettimeofday(&tv, NULL);

  if (packetTrace)
    packetTrace->addPacket(&tv, p);
  nbytes = sendto(sendSocket, p->getBuffer(), p->getIpLength(), 0, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
  if (nbytes < p->getIpLength())
    panic(ERR_CANTSEND, "PlanetLabProxy: Tried to write %d bytes raw packet, but sendto returned %d", p->getIpLength(), nbytes);

}

void PlanetLabProxy::writeToTap(Packet *p) 
{
  struct sockaddr_in sockAddr;
  int nbytes = 0;

  struct timeval tv;
  gettimeofday(&tv, NULL);
  if (packetTrace)
    packetTrace->addPacket(&tv, p);

  /* Add the Ethernet header to write directly to the TAP device */
  int len = p->getIpLength();
  for(int i = len -1; i >= 0; i--){
    p->getBuffer()[i+14] = p->getBuffer()[i];
  }
  for(int i = 0; i < 6; i++){ p->getBuffer()[i] = tapHwAddress[i]; }
  for(int i = 6; i < 11; i++){ p->getBuffer()[i] = tapHwAddress[i-6]; }
  p->getBuffer()[11] = tapHwAddress[5]+1;
  p->getBuffer()[12] = 0x08;
  p->getBuffer()[13] = 0x00;
  len += 14;

  nbytes = write(tapHandle, p->getBuffer(), len);

  if (nbytes < len)
    panic(ERR_CANTSEND, "PlanetLabProxy: Tried to write %d bytes raw packet to the TAP device, but write returned %d", len, nbytes);

}


void PlanetLabProxy::forwardPacketToReceiver(Packet *packet)
{
  /* Forward a packet to the receiver. Must set source/destination
     and TCP ports here; the policy may set them to arbitrary values! */

  if (packet->isTcpPacket()) {
    uint8 *mssOption = packet->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption && (forcedMTU >= 0)) 
      *(uint16*)&mssOption[2] = htons(forcedMTU-(sizeof(struct IpHeader)+sizeof(struct TcpHeader)));
  }
  
  packet->setSourceAndDestination(&addresses.receiverPeerAddress, &addresses.receiverAddress);
 
  writeToTap(packet);

}

void PlanetLabProxy::forwardPacketToSender(Packet *packet)
{
  /* Forward a packet to the sender. Must set source/destination
     and TCP ports here; the policy may set them to arbitrary values! */

  if (packet->isTcpPacket()) {
    uint8 *mssOption = packet->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption && (forcedMTU >= 0)) 
      *(uint16*)&mssOption[2] = htons(forcedMTU-(sizeof(struct IpHeader)+sizeof(struct TcpHeader)));
  }
  
  packet->setSourceAndDestination(&addresses.senderPeerAddress, &addresses.senderAddress);

  writeToTap(packet);

}

void PlanetLabProxy::sendPacketToRemote(Packet *packet)
{
  /* Forward a packet to the remote host. Must set source/destination
     and TCP ports here; the policy may set them to arbitrary values! 
     For PlanetLab, we have to vary the destination port for the remote 
     host. Otherwise PL will close the port if it sees a RST coming in
     instead of forwarding these packets to our slice.
  */

  assert(lastport >= 1024 && lastport <= 65534);
  addresses.remoteAddress.sin_port = htons(lastport);

  /* Increasing by just one might look like a portscan to the remote host.
   * So we take a little more effort to randomize the remote port a little. */
  //  lastport += 1;
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srandom(tv.tv_sec);
  lastport += (random() % 5) +1; // Increase by 1 to 5 port numbers

  assert(lastport >= 1024 && lastport <= 65534);
    
  /* In Planetlab, ICMP messages must contain the same Echo ID as the port
     number we bound to before. Otherwise we never see a response. */
  if(strcmp(policy->getName(), "icmp-echo") == 0) {
    struct IcmpEchoRequest *ier = (struct IcmpEchoRequest*) &(packet->getBuffer()[20]);
    struct IpHeader *ip = (struct IpHeader*) packet->getBuffer();
    uint16 iplen = packet->getIpLength();
    ier->identifier = addresses.localAddress.sin_port;
    ier->checksum = 0;
    ier->checksum = htons(Packet::checksum((uint8*)ier, iplen-20));
  }
  else if(strcmp(policy->getName(), "icmp-timestamp") == 0) {
    struct IcmpTimestampRequest *tsr = (struct IcmpTimestampRequest*) &(packet->getBuffer()[20]);
    struct IpHeader *ip = (struct IpHeader*) packet->getBuffer();
    uint16 iplen = packet->getIpLength();
    tsr->identifier = addresses.localAddress.sin_port;
    tsr->checksum = 0;
    tsr->checksum = htons(Packet::checksum((uint8*)tsr, iplen-20));
  }

  packet->setSourceAndDestination(&addresses.localAddress, &addresses.remoteAddress);
  sendRaw(packet);

}

void PlanetLabProxy::run()
{
  /* Capture packets and hand them to the policy by invoking the 
     appropriate upcall. Also, add any packets of interest to the
     packet trace. This method should never return; Monarch terminates
     via the SIGTERM handler. */

  char buf1[256], buf2[256];
  verbose(1, "Probe exchange: %s <-> %s", renderAddress(&addresses.localAddress, buf1), renderAddress(&addresses.remoteAddress, buf2));
  verbose(1, "Sender: %s <-> %s", renderAddress(&addresses.senderAddress, buf1), renderAddress(&addresses.senderPeerAddress, buf2));
  verbose(1, "Receiver: %s <-> %s", renderAddress(&addresses.receiverAddress, buf1), renderAddress(&addresses.receiverPeerAddress, buf2));

  Packet *packet = new Packet();
  while (true) {
    fd_set fdr;
    FD_ZERO(&fdr);
    FD_SET(captureSocket, &fdr);
    FD_SET(tapHandle, &fdr);

    int maxHandle = max(captureSocket, tapHandle);
    
    int retval;
    do {
      retval = select(maxHandle+1, &fdr, NULL, NULL, NULL);
      if (retval < 0) {
        if (errno == EINTR)
          continue;
        panic(ERR_SELECT, "PlanetLabProxy: Select failed (%s)", strerror(errno));
      }
    } while (false);
    
    struct timeval tv;
    gettimeofday(&tv, NULL);
    
    if (FD_ISSET(captureSocket, &fdr)) {
      int retval = read(captureSocket, packet->getBuffer(), packet->getBufferLen());
      if (retval <= 0)
        panic(ERR_CANTREAD, "PlanetLabProxy: Cannot read from capture socket: %d", retval);

      packet->setLength(retval);

      if ((packet->getSource() == ntohl(addresses.remoteAddress.sin_addr.s_addr)) && (packet->getDestination() == ntohl(addresses.localAddress.sin_addr.s_addr))) {
        if (packetTrace) 
          packetTrace->addPacket(&tv, packet);
        policy->receivePacketFromRemote(packet);
      }
    }
    
    if (FD_ISSET(tapHandle, &fdr)) {
      int retval = read(tapHandle, packet->getBuffer(), packet->getBufferLen());
      if (retval <= 0)
        panic(ERR_CANTREAD, "PlanetLabProxy: Cannot read from TAP device: %d", retval);

      /* Remove ethernet header, as we do not need or use it */
      for (int i = 14; i < retval; i++){
	packet->getBuffer()[i-14] = packet->getBuffer()[i];
      }
      packet->setLength(retval-14);

      if (packet->isSource(&addresses.senderAddress) && packet->isDestination(&addresses.senderPeerAddress)) {
        if (packetTrace)
          packetTrace->addPacket(&tv, packet);
        policy->receivePacketFromSender(packet);
      } else if (packet->isSource(&addresses.receiverAddress) && packet->isDestination(&addresses.receiverPeerAddress)) {
        if (packetTrace)
          packetTrace->addPacket(&tv, packet);
        policy->receivePacketFromReceiver(packet);
      } else {
        fprintf(stderr, "Caught an unexpeced packet on TAP (%X:%d->%X:%d); ignoring\n", 
          packet->getSource(), packet->getSourcePort(), packet->getDestination(), packet->getDestinationPort()
        );
      }
    }
  }

}
